// C Source File

#define RETURN_VALUE					// Allow a value to be returned

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200

#define ENABLE_ERROR_RETURN		// All errors to return to the OS
#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define MIN_AMS 204           // Compile for AMS 2.04 or higher
//#define SAVE_SCREEN					// Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files

static void evaluate(ESI, BOOL, BOOL);
static void equals(ESI, ESQ, HANDLE, ESI, MULTI_EXPR *, BOOL *);
static void push_with(ESI, ESI);
static MULTI_EXPR * getExpr(HANDLE);
static void writeESI(ESI, MULTI_EXPR *);
static void clean_estack();

#define NUM_VARS 4

static HANDLE variable[NUM_VARS] = {H_NULL,H_NULL,H_NULL,H_NULL};

//Main
void _main(void)
{
	ESI argptr;
	MULTI_EXPR * tempExpr;
	unsigned int x, y;
	unsigned int numArgs = ArgCount();
	BOOL approx = FALSE;
	BOOL with_expr = FALSE;
	
	for(x = 0; x < NUM_VARS; x++)
		variable[x] = H_NULL;

	//Check number of arguements
	if (numArgs < 2 || numArgs == 3)
		ER_throw(ER_TOO_FEW_ARGS);
	else if(numArgs > 4)
		ER_throw(ER_TOO_MANY_ARGS);

	//Vefify expression argument
	else if((ESTACK(EX_getArg(0)) < OR_TAG) || (ESTACK(EX_getArg(0)) > NE_TAG))
		ER_throw(ER_INVALID_SOLVE_ARG1);

	//Verify variable argument
	else if(ESTACK(EX_getArg(1)) > VAR_Q_TAG)
		ER_throw(ER_EXPECTED_VAR);

	if(!is_free_of_tag(top_estack, FLOAT_TAG) || !is_free_of_tag(NG_such_that_index, FLOAT_TAG))
		approx = TRUE;

TRY
	//Get arguments
	for(y = 0; y < 2; y++)
	{
		argptr = EX_getArg(y);
		x = argptr - next_expression_index(argptr) + 1;
		variable[y] = HeapAlloc(x * sizeof(ESQ) + sizeof(unsigned short));
		if(!variable[y])
			ER_throw(ER_MEMORY);
		tempExpr = getExpr(variable[y]);
		writeESI(argptr, tempExpr);
	}
	
	//Get bounds
	if(numArgs == 4)
	{
		if(!(is_free_of_tag(EX_getArg(2), FLOAT_TAG) && is_free_of_tag(EX_getArg(3), FLOAT_TAG)))
			approx = TRUE;
			
		argptr = EX_getArg(3);
		y = argptr - next_expression_index(argptr) + 1;
		argptr = EX_getArg(2);
		y += argptr - next_expression_index(argptr) + 1;
		
		variable[2] = HeapAlloc((2*x + y + 4) * sizeof(ESQ) + sizeof(unsigned short));
		if(!variable[2])
			ER_throw(ER_MEMORY);
			
		tempExpr = getExpr(variable[2]);
		writeESI(argptr, tempExpr);
		writeESI(HToESI(variable[1]), tempExpr);
		tempExpr->Expr[tempExpr->Size++] = GE_TAG;
		writeESI(EX_getArg(3), tempExpr);
		writeESI(HToESI(variable[1]), tempExpr);
		tempExpr->Expr[tempExpr->Size++] = LE_TAG;
		
		tempExpr->Expr[tempExpr->Size++] = AND_TAG;
		with_expr = TRUE;
		
	}

	//Get WITH expression
	if(NG_such_that_index && !with_expr)
	{
		argptr = NG_such_that_index;
		y = argptr - next_expression_index(argptr) + 1;
		variable[2] = HeapAlloc(y * sizeof(ESQ) + sizeof(unsigned short));
		if(!variable[2])
			ER_throw(ER_MEMORY);
		tempExpr = getExpr(variable[2]);
		writeESI(argptr, tempExpr);
		with_expr = TRUE;
	}

	//Create bounds if expression is trigonometric and doesn't supply bounds
	NG_approxESI(top_estack);
	if(!with_expr && !(is_free_of_tag(top_estack, SIN_TAG) && is_free_of_tag(top_estack, COS_TAG) && is_free_of_tag(top_estack, TAN_TAG)))
	{
		MO_currentOptions();
		if(ModeSettings->Angle == 1)
			ST_helpMsg("Warning: Range limited to [0,2\x8C]");
		else
			ST_helpMsg("Warning: Range limited to [0,360]");
			
		variable[2] = HeapAlloc((2*x + 13) * sizeof(ESQ) + sizeof(unsigned short));
		if(!variable[2])
			ER_throw(ER_MEMORY);
		tempExpr = getExpr(variable[2]);
		
		tempExpr->Expr[tempExpr->Size++] = 0;
		tempExpr->Expr[tempExpr->Size++] = POSINT_TAG;
		tempExpr->Expr[tempExpr->Size++] = RADIANS_TAG;
		writeESI(HToESI(variable[1]), tempExpr);
		tempExpr->Expr[tempExpr->Size++] = GE_TAG;
		
		tempExpr->Expr[tempExpr->Size++] = 2;
		tempExpr->Expr[tempExpr->Size++] = 1;
		tempExpr->Expr[tempExpr->Size++] = POSINT_TAG;
		tempExpr->Expr[tempExpr->Size++] = PI_TAG;
		tempExpr->Expr[tempExpr->Size++] = MUL_TAG;
		tempExpr->Expr[tempExpr->Size++] = RADIANS_TAG;
		writeESI(HToESI(variable[1]), tempExpr);
		tempExpr->Expr[tempExpr->Size++] = LE_TAG;
		
		tempExpr->Expr[tempExpr->Size++] = AND_TAG;
		with_expr = TRUE;
	}

	//Unlock heap space
	for(x = 1; x < NUM_VARS; x++)
		HeapUnlock(variable[x]);
		

	//Multiple expression handling
	variable[3] = HeapAlloc(sizeof(ESQ)+ sizeof(unsigned short));
	if(!variable[3])
		ER_throw(ER_MEMORY);
	tempExpr = getExpr(variable[3]);
	HeapUnlock(variable[3]);
	evaluate(HToESI(variable[0]), approx, with_expr);
	HeapUnlock(variable[0]);


	//Add with expression
	if(with_expr)
	{
		x = ((MULTI_EXPR*)HLock(variable[3]))->Size;
		HeapUnlock(variable[3]);
		y = ((MULTI_EXPR*)HLock(variable[2]))->Size;
		HeapUnlock(variable[2]);
		HANDLE tempHandle = HeapAlloc((x + y + 1) * sizeof(ESQ)+ sizeof(unsigned short));
		if(!tempHandle)
			ER_throw(ER_MEMORY);
		tempExpr = getExpr(tempHandle);
		writeESI(HToESI(variable[2]), tempExpr);
		writeESI(HToESI(variable[3]), tempExpr);
		tempExpr->Expr[tempExpr->Size++] = AND_TAG;
		HeapUnlock(variable[3]);
		HeapFreeIndir(variable+3);
		variable[3] = tempHandle;
	}


	//Final stack push
	push_simplify(HToESI(variable[3]));
	NG_approxESI(top_estack);
	NG_approxESI(HToESI(variable[3]));
	if(is_equivalent_to(top_estack, next_expression_index(top_estack)))
	{
		clean_estack();
		push_simplify(HToESI(variable[3]));
	}
	else
	{
		clean_estack();
		push_expression(HToESI(variable[3]));
	}


FINALLY
	//Clean up memory used
	for(x = 0; x < NUM_VARS; x++)
	{
		HeapUnlock(variable[x]);
		HeapFreeIndir(variable+x);
	}
ENDFINAL
}

//Processes each expression individually and writes the result to answer
void evaluate(ESI expression, BOOL approx, BOOL with_expr)
{
	ESI argptr;
	MULTI_EXPR * tempExpr;
	unsigned int x, y;
	int compare;
	ESI * solutions;
	ESQ symbol;
	HANDLE tempHandle;
	BOOL first = TRUE;
	static unsigned int total_size = 0;

	if((ESTACK(expression) == OR_TAG) || (ESTACK(expression) == AND_TAG))
	{
		evaluate(expression-1, approx, with_expr);
		evaluate(next_expression_index(expression-1), approx, with_expr);
		tempExpr = HeapDeref(variable[3]);
		tempExpr->Expr[tempExpr->Size++] = ESTACK(expression);
	}
	else
	{

		if((ESTACK(expression) < LT_TAG) || (ESTACK(expression) > NE_TAG))
			ER_throw(ER_INVALID_SOLVE_ARG1);
		else if(is_independent_of(expression, HToESI(variable[1])))
			ER_throw(ER_UNDEFINED_VAR);
		//Change the expression into left side minus right side
		symbol = ESTACK(expression);
		ESTACK(expression) = SUB_TAG;

		//Find the zeros of the equation
		if(with_expr)
			push_expression(HToESI(variable[2]));
	
		//augment(zeros(expression, variable), zeros(getDenom(expression), variable))
		push_expr2_quantum(HToESI(variable[1]), expression, GETDENOM_TAG);
		push_quantum(ZEROS_TAG);
		if(with_expr)
			push_quantum(WITH_TAG);
	
		if(with_expr)
			push_expression(HToESI(variable[2]));
		push_expr2_quantum(HToESI(variable[1]), expression, ZEROS_TAG);
		if(with_expr)
			push_quantum(WITH_TAG);
		push_quantum(AUGMENT_TAG);
		if(approx)
			push_quantum(APPROX_TAG);
		push_simplify(top_estack);
		
		if(with_expr && !(is_free_of_tag(top_estack, ARB_REAL_TAG) && is_free_of_tag(top_estack, ARB_INT_TAG)))
			ER_throw(ER_INVALID_SUCH_THAT);

		//Get the list of critical points
		y = remaining_element_count(top_estack - 1);
		if(y)
		{
			x = top_estack - next_expression_index(top_estack) + 1;
			total_size += ((*((short*)HeapDeref(variable[1])) * y) + y + x) * 3;
			tempExpr = alloca(x * sizeof(ESQ) + sizeof(unsigned short));
			tempExpr->Size = 1;
			tempExpr->Expr[0] = END_TAG;
			writeESI(top_estack, tempExpr);

			solutions = alloca(y * sizeof(ESI*));
			argptr = tempExpr->Expr + tempExpr->Size - 2;
			for(x = 0; x < y; x++)
			{
				if(ESTACK(argptr) == WHEN_TAG)
					argptr = next_expression_index(argptr-1);
				//Verify each list entry
				if(!is_constant(argptr))
					ER_throw(ER_UNDEFINED_VAR);
				solutions[x] = argptr;
				argptr = next_expression_index(argptr);
				if(ESTACK(argptr) == END_TAG)
					argptr--;
			}

			//Sort critical points using pointers to expressions (to reduce memory writes)
			for (x = 0; x < y-1; x++)
			{
				NG_approxESI(solutions[x]);
				NG_approxESI(solutions[x+1]);
				compare = compare_numbers(top_estack, next_expression_index(top_estack));
				if(compare == 0)
				{
					y--;
					argptr = solutions[x];
					solutions[x] = solutions[y];
					solutions[y] = argptr;
					x--;
				}
				else if(compare == -1)
				{
					argptr = solutions[x];
					solutions[x] = solutions[x+1];
					solutions[x+1] = argptr;
					if(x)
						x-=2;
				}
			}
			
			//Allocate space to store the answer expression
			tempHandle = HeapRealloc(variable[3], total_size * sizeof(ESQ) + sizeof(unsigned short));
			if(!tempHandle)
				ER_throw(ER_MEMORY);
			variable[3] = tempHandle;
			tempExpr = (MULTI_EXPR*)HLock(variable[3]);
			
			//Verify the lower range with the first value
			ESTACK(expression) = symbol;
			push_arg_minus_1(solutions[0]);
			push_with(expression, HToESI(variable[1]));

			NG_approxESI(top_estack);
			if(ESTACK(top_estack) == TRUE_TAG)
			{
				writeESI(solutions[0], tempExpr);
				writeESI(HToESI(variable[1]), tempExpr);
				tempExpr->Expr[tempExpr->Size++] = LT_TAG;
				if(!first)
					tempExpr->Expr[tempExpr->Size++] = OR_TAG;
				else
					first = 0;
			}

			//Verify the equality of the lower value
			equals(expression, symbol, variable[1], solutions[0], tempExpr, &first);

			for(x = 1; x < y; x++)
			{
				//Verify for each in-between value
				ESTACK(expression) = symbol;
				push_sum(solutions[x-1], solutions[x]);
				push_ushort_to_integer(2);
				push_quantum(DIV_TAG);
				push_with(expression, HToESI(variable[1]));

				NG_approxESI(top_estack);
				if(ESTACK(top_estack) == TRUE_TAG)
				{
					writeESI(solutions[x-1], tempExpr);
					writeESI(HToESI(variable[1]), tempExpr);
					tempExpr->Expr[tempExpr->Size++] = GT_TAG;
					writeESI(solutions[x], tempExpr);
					writeESI(HToESI(variable[1]), tempExpr);
					tempExpr->Expr[tempExpr->Size++] = LT_TAG;
					tempExpr->Expr[tempExpr->Size++] = AND_TAG;
					if(!first)
						tempExpr->Expr[tempExpr->Size++] = OR_TAG;
					else
						first = 0;

				}

				//Verify the equality of the upper value
				equals(expression, symbol, variable[1], solutions[x], tempExpr, &first);

			}

			//Verify the upper range with the last value
			ESTACK(expression) = symbol;
			push_arg_plus_1(solutions[y-1]);
			push_with(expression, HToESI(variable[1]));

			NG_approxESI(top_estack);
			if(ESTACK(top_estack) == TRUE_TAG)
			{
				writeESI(solutions[y-1], tempExpr);
				writeESI(HToESI(variable[1]), tempExpr);
				tempExpr->Expr[tempExpr->Size++] = GT_TAG;
				if(!first)
					tempExpr->Expr[tempExpr->Size++] = OR_TAG;
				else
					first = 0;
			}

			total_size += tempExpr->Size;

			//Unlock memory
			HeapUnlock(variable[3]);
		}

		if(first)
		{
			ESTACK(expression) = symbol;
			push1();
			push_with(expression, HToESI(variable[1]));
			push_simplify(top_estack);
			tempExpr = HeapDeref(variable[3]);
			if(ESTACK(top_estack) == TRUE_TAG)
				tempExpr->Expr[tempExpr->Size++] = TRUE_TAG;
			else
				tempExpr->Expr[tempExpr->Size++] = FALSE_TAG;
		}
	}
}

void equals(ESI expression, ESQ symbol, HANDLE variable, ESI point, MULTI_EXPR * tempExpr, BOOL * first)
{
	double compare_float;
	
	ESTACK(expression) = symbol;
	push_expression(point);
	push_with(expression, HToESI(variable));
	NG_approxESI(top_estack);
	
	if(ESTACK(top_estack) == TRUE_TAG)
	{
		writeESI(point, tempExpr);
		writeESI(HToESI(variable), tempExpr);
		tempExpr->Expr[tempExpr->Size++] = EQ_TAG;
		if(!(*first))
			tempExpr->Expr[tempExpr->Size++] = OR_TAG;
		else
			(*first) = FALSE;
	}
	else if((symbol <= GE_TAG) && (symbol >= LE_TAG))
	{
		ESTACK(expression) = SUB_TAG;
		push_expression(point);
		push_with(expression, HToESI(variable));
		push_simplify(top_estack);
		NG_approxESI(top_estack);

		if(ESTACK(top_estack) == FLOAT_TAG)
		{
			compare_float = estack_number_to_Float(top_estack);
			if((compare_float <= 5e-12) && (compare_float >= -5e-12))
			{
				writeESI(point, tempExpr);
				writeESI(HToESI(variable), tempExpr);
				tempExpr->Expr[tempExpr->Size++] = EQ_TAG;
				if(!(*first))
					tempExpr->Expr[tempExpr->Size++] = OR_TAG;
				else
					(*first) = FALSE;
			}
		}
	}
}

//Commonly used push
void push_with(ESI expression, ESI variable)
{
	push_expr_quantum(variable, EQ_TAG);
	push_expr_quantum(expression, WITH_TAG);
}

//Get MULTI_EXPR structure from a HANDLE
//Handle will be locked
MULTI_EXPR * getExpr(HANDLE source)
{
	MULTI_EXPR * tempExpr = (MULTI_EXPR*)HLock(source);
	tempExpr->Size = 1;
	tempExpr->Expr[0] = END_TAG;
	return tempExpr;
}

//Writes the expression at position source to destination
void writeESI(ESI source, MULTI_EXPR * destination)
{
	int x, y = source - next_expression_index(source);
	for(x = 0; x < y; x++)
		destination->Expr[destination->Size++] = ESTACK(source-(y-x-1));
}

//Clean up the stack
void clean_estack()
{
	while(GetArgType(top_estack) != END_TAG)
  		top_estack = next_expression_index(top_estack);
	top_estack--;
}